#ifndef GNTST_STRING_HPP
#define GNTST_STRING_HPP

#include <string>
#include <sstream>
#include <limits> // std::numeric_limits

#include "ntst_string.h"

namespace ntst
{

template <class ContainerT>
void tokenize(const std::string& str, ContainerT& tokens,
              const std::string& delimiters = " \t", bool trimEmpty = false)
    {
    typedef typename ContainerT::value_type ValueType;
    typedef typename ValueType::size_type SizeType;
    SizeType pos, lastPos = 0;
    while (true)
        {
        pos = str.find_first_of(delimiters, lastPos);
        if (pos != std::string::npos)
            {
            if (pos != lastPos || !trimEmpty)
                tokens.push_back(ValueType(str.data() + lastPos, pos - lastPos));
            }
        else
            break;
        lastPos = pos + 1;
        }
    pos = str.size();
    if (pos != lastPos || !trimEmpty)
        tokens.push_back(ValueType(str.data() + lastPos, pos - lastPos));
    }

inline bool strhasprefix(const std::string& string, const std::string& prefix)
    {
    if (string.size() < prefix.size())
        return false;
    return memcmp(string.c_str(), prefix.c_str(), prefix.size()) == 0;
    }

template <size_t arr_size>
bool strhasprefix(const char* string, char const (&prefix)[arr_size])
    {
    return (strncmp(string, prefix, arr_size - NTST_ONE_NULL_TERM) == 0);
    }

template <size_t arr_size>
const char* strnstr(const char* string, char const (&substr)[arr_size], unsigned n)
    {
    for (unsigned i = 0; i <= n; ++n)
        {
        if (strncmp(string, substr, arr_size - NTST_ONE_NULL_TERM) == 0)
            return string;
        ++string;
        }
    return NULL;
    }

/* CharPtr == "char*" || "const char*" */
template <class CharPtr, size_t ArrS>
CharPtr strstr_after(CharPtr str, char const (&substr)[ArrS])
    {
    CharPtr result = strstr(str, substr);
    if (result != NULL)
        return result + ArrS - 1;
    return NULL;
    }

template <typename StrT>
bool str_offset_to_first_not_space(const StrT& str, size_t& offset)
    {
    const size_t str_size = str.size();
    for (size_t i = offset; i < str_size; ++i)
        {
        if (!isspace(str[i]))
            {
            offset = i;
            return true;
            }
        }
    return false;
    }

inline size_t str_to_val(const char* str, int* result) {return ntst_str10toi(str, result);}
inline size_t str_to_val(const std::string& str, int* result) {return str_to_val(str.c_str(), result);}
inline size_t str_to_val(const char* str, double* result) {return ntst_strtod(str, result);}
inline size_t str_to_val(const std::string& str, double* result) {return str_to_val(str.c_str(), result);}

std::string string_printf(const char* format, ...)
#ifdef __GNUC__
__attribute__((format(printf,1,2)))
#endif
;

void ascii_string_to_lower_inplace(std::string& str);

void trim_string(std::string& str);
void trim_string(std::string& str, const char* trim);
void trim_string_left(std::string& str);
void trim_string_left(std::string& str, const char* trim);
void trim_string_right(std::string& str);
void trim_string_right(std::string& str, const char* trim);
void trim_string_trailing_decimal_zeros(std::string& str);

template <typename T>
std::string to_string_hex(T num)
    {
    std::ostringstream ss(std::ios_base::out);
    ss << std::hex << num;
    return ss.str();
    }

template <typename T>
std::string to_string_hex_pref(T num)
    {
    std::ostringstream ss(std::ios_base::out);
    ss << std::hex << "0x" << num;
    return ss.str();
    }

template <typename T>
std::wstring to_wstring_hex(T num)
    {
    std::wostringstream ss(std::ios_base::out);
    ss << std::hex << num;
    return ss.str();
    }

template <typename T>
std::wstring to_wstring_hex_pref(T num)
    {
    std::wostringstream ss(std::ios_base::out);
    ss << std::hex << L"0x" << num;
    return ss.str();
    }

#if (defined(__cplusplus) && (__cplusplus >= 201103L)) \
    || (defined(_MSC_VER) && (_MSC_VER >= 1900))
    // Not sure if _MSC_VER check is correct. to_string was bugged when appeared "ambiguous call to overloaded function". Using it with VS2015, but maybe it was already good before.
#include <type_traits>
template <typename T>
std::string to_string(T num, typename std::enable_if<!std::is_pointer<T>::value>::type* = 0)
    {
    return std::to_string(num);
    }
// For pointers
template <typename T>
std::string to_string(T num, typename std::enable_if<std::is_pointer<T>::value>::type* = 0)
    {
    return to_string_hex_pref(num);
    }
#if defined(_GLIBCXX_HAVE_BROKEN_VSWPRINTF)
#define NTST_NEED_STREAM_TO_WSTRING
#else
template <typename T>
std::wstring to_wstring(T num, typename std::enable_if<!std::is_pointer<T>::value>::type* = 0)
    {
    return std::to_wstring(num);
    }
template <typename T>
std::wstring to_wstring(T num, typename std::enable_if<std::is_pointer<T>::value>::type* = 0)
    {
    return to_wstring_hex_pref(num);
    }
#endif
#else
template <typename T>
std::string to_string(T num)
    {
    // Constructor call to pacify cppcheck
    std::ostringstream ss(std::ios_base::out);
    if (sizeof(T) == 1)
        {
         // not only char, but int8_t are also are transformed into characters
        if (std::numeric_limits<T>::is_signed)
            ss << (int)num;
        else
            ss << (unsigned)num;
        }
    else
        ss << num;
    return ss.str();
    }
#define NTST_NEED_STREAM_TO_WSTRING
#endif

#if defined(NTST_NEED_STREAM_TO_WSTRING)
template <typename T>
std::wstring to_wstring(T num)
    {
    std::wostringstream ss(std::ios_base::out);
    if (sizeof(T) == 1)
        {
        if (std::numeric_limits<T>::is_signed)
            ss << (int)num;
        else
            ss << (unsigned)num;
        }
    else
        ss << num;
    return ss.str();
    }
#undef NTST_NEED_STREAM_TO_WSTRING
#endif

template<size_t MAX_SIZE>
class NextOrMinStrFinder
    {
private:
    enum State
        {
        FIND_FIRST,
        FIND_NEXT,
        FIND_MIN
        };
public:
    NextOrMinStrFinder(const char* cur)
        {
        Init(cur);
        }
    void Init(const char* cur)
        {
        cur_str = cur;
        state = FIND_FIRST;
        }
    void PutStr(const char* new_str)
        {
        /*strcmp - Positive value if first is greater than second. */
        switch (state)
            {
            case FIND_FIRST:
                if (strcmp(new_str, cur_str) > 0)
                    state = FIND_NEXT;
                else /* new_str <= cur_str */
                    state = FIND_MIN;
                ntst_strncpy(tmp_str, new_str, MAX_SIZE);
                break;
            case FIND_NEXT:
                if (strcmp(tmp_str, new_str) > 0
                    && strcmp(new_str, cur_str) > 0)
                    {
                    /* tmp_str > new_str > cur_str
                       tmp_str = new_str */
                    ntst_strncpy(tmp_str, new_str, MAX_SIZE);
                    }
                break;
            case FIND_MIN:
                if (strcmp(new_str, cur_str) > 0)
                    {
                    /* new_str > cur_str
                       tmp_str = new_str */
                    ntst_strncpy(tmp_str, new_str, MAX_SIZE);
                    state = FIND_NEXT;
                    }
                else if (strcmp(tmp_str, new_str) > 0)
                    {
                    ntst_strncpy(tmp_str, new_str, MAX_SIZE);
                    }
                break;
            }
        }
    const char* GetStr() const
        {
        if (state != FIND_FIRST)
            return tmp_str;
        return NULL;
        }
private:
    const char* cur_str;
    State state;
    char tmp_str[MAX_SIZE];
    };

} /* namespace ntst */

#ifdef ENTST_ENABLE_QT_USE
#include <QtGlobal>
#define QStringFromStdUtf8(str) QString::fromUtf8(QByteArray::fromStdString(str))
#if (QT_VERSION < QT_VERSION_CHECK(5, 0, 0))
#include <QString>
#include <QByteArray>
static inline std::string QStringToStdUtf8(const QString& str)
    {
    QByteArray tmp = str.toUtf8();
    return std::string(tmp.constData(), tmp.size());
    }
#else
// Since qt5 toStdString() uses utf8
#define QStringToStdUtf8(str) str.toStdString()
#endif // (QT_VERSION < QT_VERSION_CHECK(5, 0, 0))
#endif // ENTST_ENABLE_QT_USE

#endif
